# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

"""Helper functions to work with generators"""

import time
import queue
import threading
from typing import Iterable, Generator


def chunk(g: Iterable, size=10000):
    """Formats output of iterable into chunks of equal sizes

    Arguments:
        g {Iterable} -- An iterable preferrably a generator

    Keyword Arguments:
        size {int} -- The size of each chunk (default: {10000})

    Yields:
        List[object] -- A list of items from `g`
    """
    chunk = []
    for item in g:
        chunk.append(item)
        if len(chunk) >= size:
            yield chunk
            chunk = []
    if len(chunk) > 0:
        yield chunk


def prefetch(g: Iterable, size=1) -> Generator:
    """Prefetch wrapper around a generator/iterable

    Arguments:
        g {Iterable} -- An iterable preferably a generator

    Keyword Arguments:
        size {int} -- The size of the prefetch queue (default: {1})

    Yields:
        object -- items from `g`
    """
    coord = {'done': False}
    prefetch_queue = queue.Queue(size)

    def fill_prefetch_queue():
        for item in g:
            while not coord['done']:
                try:
                    prefetch_queue.put(item, block=True, timeout=0.1)
                except queue.Full:
                    continue
                else:
                    break

            if coord['done']:
                break

        # Wait for prefetch_queue to be completely consumed
        while prefetch_queue.qsize() > 0:
            time.sleep(0.1)
            continue
        coord['done'] = True

    worker = threading.Thread(target=fill_prefetch_queue)
    worker.setDaemon(True)
    worker.start()

    while not coord['done']:
        try:
            item = prefetch_queue.get(block=True, timeout=0.1)
        except queue.Empty:
            continue
        else:
            yield item

    worker.join()



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

"""Convenient interface to perform sentence encoding"""

from typing import Optional, List

import numpy as np

import torch
from torch import nn

from laser.encoder import load_encoder_from_file
from laser.data import Batcher


class SentenceEncoder(nn.Module):

    def __init__(
        self,
        lang: str,
        model_path: str,
        bpe_codes: str,
        max_seq_length: int = 256,
        max_sents: Optional[int] = None,
        max_tokens: Optional[int] = 12000,
    ):
        """Wrapper around Encoder and Batcher to encode sentences

        Arguments:
            lang {str} -- Language expected
                (this would only affect preprocessing)
            model_path {str} -- Path to model file
            bpe_codes {str} -- Path to bpe codes file

        Keyword Arguments:
            max_seq_length {int} --
                Maximum number of tokens per sentence (default: {256})
            max_sents {Optional[int]} --
                Maximum number of sentences per batch (default: {None})
            max_tokens {Optional[int]} --
                Maximum number of tokens per batch (default: {12000})
        """
        super().__init__()
        self.encoder, dictionary = load_encoder_from_file(model_path)
        self.batcher = Batcher(
            lang=lang,
            bpe_codes=bpe_codes,
            dictionary=dictionary,
            max_seq_length=max_seq_length,
            max_sents=max_sents,
            max_tokens=max_tokens,
        )

        # Default load in eval mode
        self.eval()

    @torch.no_grad()
    def encode_sentences(self, texts: List[str]) -> np.ndarray:
        """Encodes a list of sentences

        Arguments:
            texts {List[str]} -- A list of raw texts

        Returns:
            np.ndarray -- Sentence embeddings
        """
        device = self.encoder.embed_tokens.weight.device

        embeddings = [None] * len(texts)
        for batch in self.batcher.make_batches(texts):
            outputs = self.encoder(
                batch.tokens.to(device),
                batch.lengths.to(device),
            ).sentemb.detach().cpu().numpy()

            for i, embedding in zip(batch.indices, outputs):
                embeddings[i] = embedding

        return np.stack(embeddings, axis=0)

    def forward(self, *args, **kwargs):
        """Alias for encode_sentences"""
        return self.encode_sentences(*args, **kwargs)



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

"""Misc helper functions"""

import torch


def open_text_file(filepath: str, mode: str = 'r'):
    return open(filepath, mode=mode, encoding='utf-8', newline='\n')


def determine_device(use_gpu):
    if use_gpu and torch.cuda.is_available():
        return torch.device('cuda:0')
    return torch.device('cpu')



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

"""Data processing and loading classes and helper functions"""

import os
import warnings
from typing import NamedTuple, Iterator, Generator, List, Optional

import torch
import numpy as np


MOSES_CHUNK_SIZE = 128


class Tokenizer(object):

    def __init__(self, lang: str = 'en'):
        """Create a tokenizer instance for a specific language.
        Take note that a caveat is that this tokenizer is not picklable.

        Does the following in sequence
        1. Punct normalize
        2. Lang specific moses tokenize
        3. Lang specific addtional segmentation

        Keyword Arguments:
            lang {str} -- Language identifier (default: {'en'})

        Raises:
            ImportError: Requires sacremoses
        """
        try:
            from mosestokenizer import \
                MosesPunctuationNormalizer, MosesTokenizer
        except ImportError:
            raise ImportError('Please install sacremoses')

        self.lang = lang
        self.punct_normalizer = MosesPunctuationNormalizer(lang=lang)
        self.tokenizer = MosesTokenizer(lang=lang)

        if self.lang == 'zh':
            import opencc
            self.opencc_converter = opencc.OpenCC('tw2s')

            import jieba
            self.jieba_tokenizer = jieba.dt

        elif self.lang == 'jp':
            import MeCab
            self.wakati = MeCab.Tagger('-Owakati')

        elif self.lang == 'el':
            import transliterate
            self.translit = transliterate.translit

    def tokenize(self, texts: List[str]) -> List[str]:
        """Tokenizes a string. Output is string rather than list of tokens.

        Arguments:
            text {List[str]} -- Raw input texts

        Returns:
            List[str] -- Tokenized sentences
        """
        texts = [' '.join(t.lower().split()) for t in texts]
        tokenized_texts = []
        for i in range(0, len(texts), MOSES_CHUNK_SIZE):
            chunk = texts[i:(i + MOSES_CHUNK_SIZE)]

            # Normalize punctuations
            [self.punct_normalizer.writeline(t) for t in chunk]

            # Tokenize
            [self.tokenizer.writeline(
                self.punct_normalizer.readline()
            ) for _ in chunk]

            tokenized_texts += [self.tokenizer.readline() for _ in chunk]
            # This may seem like a lot of for loops
            # but the overhead of tokenization should be much higher
            # Also this is an efficient way to buffer I/O to/from the
            # perl scripts

            # Random note: it might be possible to run these piped processes
            # on a separate thread.
            # Would not improve the current performance of this script by
            # much but still a cool experiment.

        if self.lang == 'zh':
            texts = [
                ' '.join(self.jieba_tokenizer.cut(
                    self.opencc_converter.convert(text),
                    cut_all=False, HMM=True
                ))
                for text in texts
            ]

        elif self.lang == 'jp':
            texts = [
                self.wakati.parse(text) if len(text) > 0 else ''
                for text in texts
            ]

        elif self.lang == 'el':
            texts = [
                self.translit(text, language_code='el')
                for text in texts
            ]

        return texts


class Batch(NamedTuple):
    indices: List[int]
    texts: List[str]
    tokens: torch.Tensor
    lengths: torch.Tensor


class Batcher(object):

    def __init__(
        self,
        lang: str,
        bpe_codes: str,
        dictionary: dict,
        max_seq_length: int = 256,
        max_sents: Optional[int] = None,
        max_tokens: Optional[int] = 12000
    ):
        self.lang = lang

        # Load tokenizer and bpe model
        try:
            from fastBPE import fastBPE
        except ImportError:
            raise ImportError('Please install fastBPE first')
        assert os.path.isfile(bpe_codes)

        self.tokenizer = Tokenizer(lang=lang)
        self.bpe_model = fastBPE(bpe_codes)

        # Set dictionary
        assert (
            '<pad>' in dictionary and
            '</s>' in dictionary and
            '<unk>' in dictionary
        ), 'dictionary appears to be invalid'

        self.dictionary = dictionary
        self.pad_index = self.dictionary['<pad>']
        self.eos_index = self.dictionary['</s>']
        self.unk_index = self.dictionary['<unk>']

        # Setup max_seq_length, max_sents and max_tokens
        if max_sents is None and max_tokens is None:
            warnings.warn(
                'Using batcher without specifying max_sents and max_tokens '
                'defaulting to using max_sents of 1. '
                'This can be highly inefficient.'
            )
            max_sents = 1

        self.max_seq_length = max_seq_length
        self.max_sents = max_sents
        self.max_tokens = max_tokens

    def _text_to_token_ids(self, text: str) -> List[int]:
        """Convert a text into a list of token ids
        Sentence that is too long will be truncated

        Arguments:
            text {str} -- BPE encoded text

        Returns:
            List[int] -- List of token ids
        """
        tokens = [self.dictionary.get(t, self.unk_index) for t in text.split()]
        if len(tokens) > self.max_seq_length - 1:
            warnings.warn((
                'Found sentence with over {} tokens '
                'sentence will be truncated to a valid length'
            ).format(self.max_seq_length))
            tokens = tokens[:(self.max_seq_length - 1)]
        tokens.append(self.eos_index)
        return tokens

    def _process_texts(self, texts: List[str]) -> List[List[int]]:
        """Process a list of texts

        Arguments:
            texts {List[str]} -- Raw texts

        Returns:
            List[List[int]] -- List of token ids
        """
        texts = self.tokenizer.tokenize(texts)
        texts = self.bpe_model.apply(texts)
        return [self._text_to_token_ids(text) for text in texts]

    def _collate_tokens(self, tokens: List[List[int]]) -> torch.Tensor:
        """Right pad a list of token ids

        Arguments:
            tokens {List[List[int]]} -- List of token ids

        Returns:
            torch.Tensor -- A 2D tensor of shape [batch_size, seq_len]
        """
        batch_length = max([len(t) for t in tokens])
        batch_length = min(batch_length, self.max_seq_length)

        def pad_toks(toks: List[int]) -> List[int]:
            length = len(toks)
            if length < batch_length:
                toks += [self.pad_index] * (batch_length - length)
            return toks

        return torch.LongTensor([pad_toks(t) for t in tokens])

    def make_batches(
        self, texts: Iterator[str]
    ) -> Generator[Batch, None, None]:
        """Create batches from raw texts
        Sequences in each batch will be sorted based on longest to shortest.
        Batches will also be right padded.

        TODO: Should warn if too many unknown tokens are produced

        Arguments:
            texts {Iterator[str]} -- A list of raw texts

        Yields:
            List[int] -- batch_indices
                The original indices of input text
            List[str] -- batch_texts
                The input texts for this batch
            torch.Tensor -- batch_tokens
                The token_ids padded into a 2D tensor of shape
                [batch_size, seq_len]
            torch.Tensor -- batch_lengths
                The unpadded sequence lengths of each entry.
                A 1D tensor of shape [batch_size]
        """
        tokens = self._process_texts(texts)
        lengths = [len(t) for t in tokens]
        indices = np.argsort(lengths)[::-1]

        batch_indices = []
        batch_texts = []
        batch_tokens = []
        max_batch_length = 0
        for i in indices:
            if (
                (
                    self.max_sents is not None and
                    len(batch_tokens) >= self.max_sents
                ) or (
                    self.max_tokens is not None and (
                        max(max_batch_length, lengths[i]) *
                        (len(batch_tokens) + 1) > self.max_tokens
                    )
                )
            ):
                # yield cur batch
                batch_tokens = self._collate_tokens(batch_tokens)
                yield Batch(
                    indices=batch_indices,
                    texts=batch_texts,
                    tokens=batch_tokens,
                    lengths=torch.LongTensor([len(t) for t in batch_tokens])
                )
                batch_indices = []
                batch_texts = []
                batch_tokens = []
                max_batch_length = 0

            batch_indices.append(i)
            batch_texts.append(texts[i])
            batch_tokens.append(tokens[i])
            max_batch_length = max(max_batch_length, lengths[i])

        if len(batch_tokens) > 0:
            batch_tokens = self._collate_tokens(batch_tokens)
            yield Batch(
                indices=batch_indices,
                texts=batch_texts,
                tokens=batch_tokens,
                lengths=torch.LongTensor([len(t) for t in batch_tokens])
            )



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

"""Download a pretrained model distributed by
the authors of the LASER project"""

import os
import subprocess
from typing import Tuple


S3_URL = 'https://dl.fbaipublicfiles.com/laser/models'
PRETRAINED_CODES_FILE = '93langs.fcodes'
PRETRAINED_MODEL_FILE = 'bilstm.93langs.2018-12-26.pt'
CACHE_DIR = os.path.join(os.environ['HOME'], '.cache/laser')


def download_file(url: str, filepath: str, force: bool = False):
    if not os.path.isfile(filepath) or force:
        assert subprocess.call('command -v curl', shell=True) == 0, \
            '`curl` is needed to download pretrained models'

        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        cmd = 'curl {} -o {}'.format(url, filepath)
        assert subprocess.call(cmd, shell=True) == 0, \
            'Failed to download from {}'.format(url)


def get_pretrained_model_paths(cache_dir: str = CACHE_DIR) -> Tuple[str, str]:
    """Download and cache the pretrained models and return the model save
    paths. On future calls, downloading will be skipped.

    Keyword Arguments:
        cache_dir {str} -- Directory to cache models (default: {CACHE_DIR})

    Returns:
        Tuple[str, str] -- (bpe_codes file, encoder model file)
    """
    for filename in [PRETRAINED_CODES_FILE, PRETRAINED_MODEL_FILE]:
        url = '{}/{}'.format(S3_URL, filename)
        filepath = os.path.join(CACHE_DIR, 'models', filename)
        download_file(url=url, filepath=filepath)

    return (
        os.path.join(CACHE_DIR, 'models', PRETRAINED_CODES_FILE),
        os.path.join(CACHE_DIR, 'models', PRETRAINED_MODEL_FILE)
    )



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

from laser.data import Batch, Batcher
from laser.encoder import EncoderOuts, Encoder
from laser.sent_encoder import SentenceEncoder

from laser.encoder import load_encoder_from_file

__all__ = [
    'Batch',
    'Batcher',
    'EncoderOuts',
    'Encoder',
    'SentenceEncoder',
    'load_encoder_from_file',
]

__version__ = '1.0.0'



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

from laser.bin import encode_texts, filter_parallel

__all__ = ['encode_texts', 'filter_parallel']



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

"""LASER LSTM sequence encoder"""

import torch
from torch import nn
from typing import NamedTuple, Optional, Tuple


class EncoderOuts(NamedTuple):
    sentemb: torch.Tensor
    encoder_out: Optional[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]
    encoder_padding_mask: Optional[torch.Tensor]


class Encoder(nn.Module):

    def __init__(
        self,
        num_embeddings: int,
        padding_idx: int,
        embed_dim: int = 320,
        hidden_size: int = 512,
        num_layers: int = 1,
        bidirectional: bool = False,
        padding_value: float = 0.0
    ):
        """LSTM based sequence encoder

        Arguments:
            num_embeddings {int} -- Number of unique token embeddings
            padding_idx {int} -- Padding token id

        Keyword Arguments:
            embed_dim {int} -- Embedding dimension size (default: {320})
            hidden_size {int} -- Hidden layer dimension size (default: {512})
            num_layers {int} -- Number of LSTM layers (default: {1})
            bidirectional {bool} -- Use Bidirectional LSTM (default: {False})
            padding_value {float} --
                Value to pad hidden layers (default: {0.0})
        """
        super().__init__()

        self.num_embeddings = int(num_embeddings)
        self.padding_idx = int(padding_idx)
        self.embed_dim = int(embed_dim)
        self.hidden_size = int(hidden_size)
        self.num_layers = int(num_layers)
        self.bidirectional = bool(bidirectional)
        self.padding_value = float(padding_value)

        self.embed_tokens = nn.Embedding(
            num_embeddings=self.num_embeddings,
            embedding_dim=self.embed_dim,
            padding_idx=self.padding_idx
        )

        self.lstm = nn.LSTM(
            input_size=self.embed_dim,
            hidden_size=self.hidden_size,
            num_layers=self.num_layers,
            bidirectional=self.bidirectional
        )

        self.num_directions = 2 if self.bidirectional else 1
        self.output_units = self.num_directions * self.hidden_size

    def forward(
        self,
        src_tokens: torch.LongTensor,
        src_lengths: torch.LongTensor,
        return_encoder_out: bool = False,
        return_encoder_padding_mask: bool = False,
    ) -> EncoderOuts:
        """Encode a batch of sequences

        Arguments:
            src_tokens {torch.LongTensor} -- [batch_size, seq_len]
            src_lengths {torch.LongTensor} -- [batch_size]

        Keyword Arguments:
            return_encoder_out {bool} --
                Return output tensors? (default: {False})
            return_encoder_padding_mask {bool} --
                Return encoder padding mask? (default: {False})

        Returns:
            [type] -- [description]
        """
        bsz, seqlen = src_tokens.size()

        x = self.embed_tokens(src_tokens)
        x = x.transpose(0, 1)  # BTC -> TBC

        # Pack then apply LSTM
        packed_x = nn.utils.rnn.pack_padded_sequence(
            x, src_lengths, batch_first=False, enforce_sorted=True)
        packed_outs, (final_hiddens, final_cells) = \
            self.lstm.forward(packed_x)

        x, _ = nn.utils.rnn.pad_packed_sequence(
            packed_outs, padding_value=self.padding_value)
        assert list(x.size()) == [seqlen, bsz, self.output_units]

        # Set padded outputs to -inf so they are not selected by max-pooling
        padding_mask = src_tokens.eq(self.padding_idx).t()
        if padding_mask.any():
            x = x.float().masked_fill_(
                mask=padding_mask.unsqueeze(-1),
                value=float('-inf'),
            ).type_as(x)

        # Build the sentence embedding by max-pooling over the encoder outputs
        sentemb = x.max(dim=0)[0]

        encoder_out = None
        if return_encoder_out:
            final_hiddens = self._combine_outs(final_hiddens)
            final_cells = self._combine_outs(final_cells)
            encoder_out = (x, final_hiddens, final_cells)

        encoder_padding_mask = None
        if return_encoder_padding_mask:
            encoder_padding_mask = src_tokens.eq(self.padding_idx).t()

        return EncoderOuts(
            sentemb=sentemb,
            encoder_out=encoder_out,
            encoder_padding_mask=encoder_padding_mask
        )

    def _combine_outs(self, x: torch.Tensor) -> torch.Tensor:
        """Combines outputs for same layer same entry in batch.
        Only used for BiLSTM.

        Arguments:
            outs {torch.Tensor} -- [num_layers * num_dir, bsz, hidden_size]

        Returns:
            torch.Tensor -- [num_layers, bsz, num_dir * hidden_size]
        """
        # [num_layers * num_dir, bsz, hidden_size]
        #   -> [num_layers, num_dir, bsz, hidden_size]
        #   -> [num_layers, bsz, num_dir, hidden_size]
        #   -> [num_layers, bsz, num_dir * hidden_size]
        if self.num_directions == 1:
            return x
        x = x.reshape(
            self.num_layers,
            self.num_directions,
            -1,
            self.hidden_size
        ).transpose(1, 2)
        x = x.transpose(1, 2)
        return x.reshape(self.num_layers, -1, self.output_units)


def load_encoder_from_file(model_path: str) -> Tuple[Encoder, dict]:
    # Load state_dict
    state_dict = torch.load(model_path, map_location='cpu')
    if 'left_pad' in state_dict['params']:
        del state_dict['params']['left_pad']

    # Create encoder
    encoder = Encoder(**state_dict['params'])
    encoder.load_state_dict(state_dict['model'])
    encoder.eval()

    return encoder, state_dict['dictionary']



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

"""Filter a parallel corpus"""

import os
import argparse
from typing import List, Tuple

import threading

from time import time
from tqdm import tqdm

import torch
import numpy as np

from laser.data import Batcher, Batch
from laser.encoder import load_encoder_from_file
from laser.utils import open_text_file, determine_device
from laser import generator_utils as gen_utils
from laser.pretrained import CACHE_DIR, get_pretrained_model_paths

CACHE = {}


def add_options(parser: argparse.ArgumentParser):
    # I/O options
    io_group = parser.add_argument_group('I/O options')
    io_group.add_argument(
        'prefixes', metavar='PREFS', nargs='+', type=str,
        help='Parallel corpus file prefixes')
    io_group.add_argument(
        '-o', '--output', metavar='PREF', type=str, required=True,
        help='Prefix to a file to store outputs')

    # Model options
    model_group = parser.add_argument_group('Model options')
    model_group.add_argument(
        '-l', '--langs', metavar='LANG', type=str, nargs=2, required=True,
        help='Source and target langauge identifier')
    model_group.add_argument(
        '-c', '--cache-dir', metavar='DIR', type=str, default=CACHE_DIR,
        help='Path to the directory where model is cached')
    model_group.add_argument(
        '-b', '--bpe-codes', metavar='FILE', type=str,
        help='Path to the bpe codes')
    model_group.add_argument(
        '-m', '--model', metavar='FILE', type=str,
        help='Path to the LASER bilstm model')

    # Filter options
    filter_group = parser.add_argument_group('Filter options')
    filter_group.add_argument(
        '--threshold', metavar='FLOAT', type=float, default=1.04,
        help='L2 distance threshold to filter sentence pairs')
    filter_group.add_argument(
        '--output-filtered', action='store_true',
        help='Write filtered data to <output_file>.filtered')

    # Inference options
    inference_group = parser.add_argument_group('Inference options')
    inference_group.add_argument(
        '--max-seq-len', metavar='INT', type=int, default=256,
        help='Maximum number of tokens per sentence')
    inference_group.add_argument(
        '--max-sents', metavar='INT', type=int,
        help='Maximum number sentences per batch')
    inference_group.add_argument(
        '--max-tokens', metavar='INT', type=int, default=12000,
        help='Maximum number tokens per batch')
    inference_group.add_argument(
        '--gpu', action='store_true',
        help='Run model with CPU?')
    inference_group.add_argument(
        '--fp16', action='store_true',
        help='Run model with FP16?')

    # Data loading options
    data_group = parser.add_argument_group('Data loading options')
    data_group.add_argument(
        '--chunk-size', metavar='INT', type=int, default=10000,
        help='The chunk size that input files should be read in')
    data_group.add_argument(
        '--num-workers', metavar='INT', type=int, default=4,
        help='Number of workers to use to generate new batches')

    return parser


def download_pretrained_models(args: argparse.Namespace):
    if args.bpe_codes is None:
        args.bpe_codes = get_pretrained_model_paths(args.cache_dir)[0]
    if args.model is None:
        args.model = get_pretrained_model_paths(args.cache_dir)[1]


def create_batchers(
    dictionary: dict,
    args: argparse.Namespace,
) -> Tuple[Batcher, Batcher]:
    kwargs = dict(
        bpe_codes=args.bpe_codes,
        dictionary=dictionary,
        max_seq_length=args.max_seq_len,
        max_sents=args.max_sents,
        max_tokens=args.max_tokens,
    )
    src_batcher = Batcher(lang=args.langs[0], **kwargs)
    tgt_batcher = Batcher(lang=args.langs[1], **kwargs)
    return src_batcher, tgt_batcher


def create_input_stream(args: argparse.Namespace):
    """Reads all files referenced by prefixes and langs

    Arguments:
        args {argparse.Namespace} -- Parsed command line options

    Yields:
        Tuple[str, str] -- src_line, tgt_line
    """
    for pref in args.prefixes:
        src_filepath = '{}.{}'.format(pref, args.langs[0])
        tgt_filepath = '{}.{}'.format(pref, args.langs[1])

        with open_text_file(src_filepath, 'r') as fsrc:
            with open_text_file(tgt_filepath, 'r') as ftgt:
                for src_line, tgt_line in zip(fsrc, ftgt):
                    yield src_line, tgt_line


def create_batches(
    params: Tuple[
        List[Tuple[str, str]],
        dict,
        argparse.Namespace
    ]
) -> Tuple[List[Batch], List[Batch]]:
    chunk, dictionary, args = params
    src_batcher, tgt_batcher = load_batchers(dictionary, args)
    src_lines, tgt_lines = zip(*chunk)
    src_batches = list(src_batcher.make_batches(src_lines))
    tgt_batches = list(tgt_batcher.make_batches(tgt_lines))
    return src_batches, tgt_batches


def load_batchers(
    dictionary: dict,
    args: argparse.Namespace,
) -> Tuple[Batcher, Batcher]:
    if 'src_batcher' not in CACHE:
        CACHE['src_batcher'], CACHE['tgt_batcher'] = \
            create_batchers(dictionary, args)
    return CACHE['src_batcher'], CACHE['tgt_batcher']


def init_output_files(args: argparse.Namespace):
    output_pref = os.path.abspath(args.output)
    os.makedirs(os.path.dirname(output_pref), exist_ok=True)

    src_outpath = '{}.{}'.format(output_pref, args.langs[0])
    tgt_outpath = '{}.{}'.format(output_pref, args.langs[1])

    with open_text_file(src_outpath, 'w'):
        pass
    with open_text_file(tgt_outpath, 'w'):
        pass

    if args.output_filtered:
        with open_text_file(src_outpath + '.filtered', 'w'):
            pass
        with open_text_file(tgt_outpath + '.filtered', 'w'):
            pass


def append_results_to_file(
    src_texts: List[str],
    src_embeddings: np.ndarray,
    tgt_texts: List[str],
    tgt_embeddings: np.ndarray,
    args: argparse.Namespace,
):
    assert len(src_texts) == len(src_embeddings)
    assert len(tgt_texts) == len(tgt_embeddings)
    assert len(src_texts) == len(tgt_texts)

    src_outpath = '{}.{}'.format(args.output, args.langs[0])
    tgt_outpath = '{}.{}'.format(args.output, args.langs[1])
    if src_embeddings.dtype != np.float32:
        src_embeddings = src_embeddings.astype(np.float32)
    if tgt_embeddings.dtype != np.float32:
        tgt_embeddings = tgt_embeddings.astype(np.float32)
    l2_dists = np.sqrt(np.sum((src_embeddings - tgt_embeddings) ** 2, axis=1))

    with open_text_file(src_outpath, 'a') as fsrc:
        with open_text_file(tgt_outpath, 'a') as ftgt:
            for i, d in enumerate(l2_dists):
                if d <= args.threshold:
                    fsrc.write(src_texts[i])
                    ftgt.write(tgt_texts[i])

    if not args.output_filtered:
        return

    with open_text_file(src_outpath + '.filtered', 'a') as fsrc:
        with open_text_file(tgt_outpath + '.filtered', 'a') as ftgt:
            for i, d in enumerate(l2_dists):
                if d > args.threshold:
                    fsrc.write(src_texts[i])
                    ftgt.write(tgt_texts[i])


def main(args):
    if args.num_workers >= 1:
        import multiprocessing
        num_workers = args.num_workers
    else:
        from multiprocessing import dummy as multiprocessing
        num_workers = 1

    # Download pretrained model if needed
    download_pretrained_models(args)

    start_time = time()

    # Load encoder
    device = determine_device(args.gpu)
    encoder, dictionary = load_encoder_from_file(args.model)
    encoder = encoder.eval()
    encoder = encoder.to(device)
    if args.fp16:
        encoder = encoder.half()

    # Define create_loader fn
    semaphore = threading.Semaphore(num_workers)

    def create_loader():
        for chunk in gen_utils.chunk(
            create_input_stream(args),
            size=args.chunk_size,
        ):
            semaphore.acquire()
            yield chunk, dictionary, args

    # Define process_batches fn
    @torch.no_grad()
    def process_batches(batches: List[Batch]) -> Tuple[List[str], np.ndarray]:
        chunk_size = sum([len(batch[0]) for batch in batches])
        chunk_texts = [None] * chunk_size
        chunk_embeddings = [None] * chunk_size
        for b in batches:
            outputs = encoder.forward(
                b.tokens.to(device),
                b.lengths.to(device),
            ).sentemb
            outputs = outputs.detach().cpu().numpy()

            for i, text, embedding in zip(b.indices, b.texts, outputs):
                chunk_texts[i] = text
                chunk_embeddings[i] = embedding

        return chunk_texts, np.array(chunk_embeddings)

    init_output_files(args)

    # Loop through each chunk
    # 1. Create batches (on subprocesses)
    # 2. Inference with bilstm
    # 3. Save results to file
    with tqdm(unit='sent') as pbar:
        with multiprocessing.Pool(num_workers) as pool:
            batches_generator = pool.imap(create_batches, create_loader())

            t0 = time()
            for src_batches, tgt_batches in batches_generator:
                semaphore.release()

                t1 = time()
                src_texts, src_embeddings = process_batches(src_batches)
                tgt_texts, tgt_embeddings = process_batches(tgt_batches)

                t2 = time()
                torch.cuda.empty_cache()
                append_results_to_file(
                    src_texts, src_embeddings,
                    tgt_texts, tgt_embeddings,
                    args,
                )

                t3 = time()
                pbar.set_postfix({
                    'data_time': t1 - t0,
                    'inference_time': t2 - t1,
                    'write_time': t3 - t2,
                })
                pbar.update(len(src_texts))
                t0 = time()

    time_taken = time() - start_time
    print('Finished in {:.1f}s'.format(time_taken))



import os
import re
import ast
import setuptools

_this_dir = os.path.abspath(os.path.dirname(__file__))

# Parse version from ./laser/__init__.py
_version_re = re.compile(r'__version__\s+=\s+(.*)')
_init_file = os.path.join(_this_dir, 'laser', '__init__.py')
with open(_init_file, 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))


# Parse requirements form ./requirements.txt
_requirements_file = os.path.join(_this_dir, 'requirements.txt')
with open(_requirements_file, 'rb') as f:
    install_requires = [l.decode('utf-8').strip() for l in f]


# Parse long description from ./README.md
_readme_file = os.path.join(_this_dir, 'README.md')
with open(_readme_file, 'rb') as f:
    long_description = f.read().decode('utf-8')


setuptools.setup(
    # General information
    name='laser-keep-alive',
    version=version,

    description='Keeping the original LASER project alive',
    long_description=long_description,
    long_description_content_type='text/markdown',

    author='Wang Ming Rui',
    author_email='mingruimingrui@hotmail.com',
    url='https://github.com/mingruimingrui/laser-keep-alive',

    # Build arguments
    packages=setuptools.find_packages(exclude=['models', 'tests']),
    python_requires='>=3.6.1',
    install_requires=install_requires,

    # PyPI package information
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='laser sentence embedding',
    license='MIT',
)



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
from laser import bin as laser_bin


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog='python -m laser',
        description='Language-Agnostic SEntence Representations',
    )
    subparsers = parser.add_subparsers(dest='subcommand')

    encode_parser = subparsers.add_parser(
        'encode', help='Encode a text file line by line')
    laser_bin.encode_texts.add_options(encode_parser)

    filter_parser = subparsers.add_parser(
        'filter', help='Filter a parallel corpus based on similarity')
    laser_bin.filter_parallel.add_options(filter_parser)

    return parser


if __name__ == "__main__":
    parser = make_parser()
    args = parser.parse_args()

    if args.subcommand == 'encode':
        laser_bin.encode_texts.main(args)

    elif args.subcommand == 'filter':
        laser_bin.filter_parallel.main(args)

    else:
        parser.print_help()



# Copyright (c) 2020 mingruimingrui
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the MIT-style license found in the
# LICENSE file in the root directory of this source tree.

"""Encode a text file line by line"""

import os
import io
import argparse
from typing import Generator, List, Tuple

import threading

from time import time
from tqdm import tqdm

import torch
import numpy as np

from laser.data import Batcher, Batch
from laser.encoder import load_encoder_from_file
from laser.utils import open_text_file, determine_device
from laser import generator_utils as gen_utils
from laser.pretrained import CACHE_DIR, get_pretrained_model_paths

CACHE = {}


def add_options(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Adds options for this script to an existing parser
    options will be added inplace so there isn't a need to use the parser
    returned by this function.

    Arguments:
        parser {argparse.ArgumentParser} -- A given parser

    Returns:
        argparse.ArgumentParser -- The input parser
    """
    # I/O options
    io_group = parser.add_argument_group('I/O options')
    io_group.add_argument(
        'files', metavar='FILE', nargs='+', type=str,
        help='Text files to encode')
    io_group.add_argument(
        '-o', '--output', metavar='PREF', type=str, required=True,
        help='File to store outputs')
    io_group.add_argument(
        '--output-format', metavar='FORMAT', type=str,
        choices=['npy', 'csv'], default='npy',
        help='The output format')
    io_group.add_argument(
        '--output-fp16', action='store_true',
        help='Output results as fp16')

    # Model options
    model_group = parser.add_argument_group('Model options')
    model_group.add_argument(
        '-l', '--lang', metavar='LANG', type=str, required=True,
        help='Langauge identifier')
    model_group.add_argument(
        '-c', '--cache-dir', metavar='DIR', type=str, default=CACHE_DIR,
        help='Path to the directory where model is cached')
    model_group.add_argument(
        '-b', '--bpe-codes', metavar='FILE', type=str,
        help='Path to the bpe codes')
    model_group.add_argument(
        '-m', '--model', metavar='FILE', type=str,
        help='Path to the LASER bilstm model')

    # Inference options
    inference_group = parser.add_argument_group('Inference options')
    inference_group.add_argument(
        '--max-seq-len', metavar='INT', type=int, default=256,
        help='Maximum number of tokens per sentence')
    inference_group.add_argument(
        '--max-sents', metavar='INT', type=int,
        help='Maximum number sentences per batch')
    inference_group.add_argument(
        '--max-tokens', metavar='INT', type=int, default=12000,
        help='Maximum number tokens per batch')
    inference_group.add_argument(
        '--gpu', action='store_true',
        help='Run model with CPU?')
    inference_group.add_argument(
        '--fp16', action='store_true',
        help='Run model with FP16?')

    # Data loading options
    data_group = parser.add_argument_group('Data loading options')
    data_group.add_argument(
        '--chunk-size', metavar='INT', type=int, default=10000,
        help='The chunk size that input files should be read in')
    data_group.add_argument(
        '--num-workers', metavar='INT', type=int, default=4,
        help='Number of workers to use to generate new batches')

    return parser


def download_pretrained_models(args: argparse.Namespace):
    if args.bpe_codes is None:
        args.bpe_codes = get_pretrained_model_paths(args.cache_dir)[0]
    if args.model is None:
        args.model = get_pretrained_model_paths(args.cache_dir)[1]


def create_batcher(dictionary: dict, args: argparse.Namespace) -> Batcher:
    return Batcher(
        lang=args.lang,
        bpe_codes=args.bpe_codes,
        dictionary=dictionary,
        max_seq_length=args.max_seq_len,
        max_sents=args.max_sents,
        max_tokens=args.max_tokens,
    )


def load_batcher(dictionary: dict, args: argparse.Namespace) -> Batcher:
    """Loads batcher into CACHE and on subsequent calls, retrieves batcher
    from cache.

    Arguments:
        dictionary {dict} -- Batcher dictionary
        args {argparse.Namespace} -- Parsed commandline options

    Returns:
        Batcher -- Text batcher
    """
    if 'batcher' not in CACHE:
        CACHE['batcher'] = create_batcher(dictionary, args)
    return CACHE['batcher']


def create_input_stream(
    args: argparse.Namespace
) -> Generator[str, None, None]:
    """Create a generator that loads lines from text files

    Arguments:
        args {argparse.Namespace} -- Parsed commandline options

    Yields:
        str -- Lines from input files
    """
    for filepath in args.files:
        with open_text_file(filepath, 'r') as f:
            for line in f:
                yield line


def create_batches(
    params: Tuple[
        List[str],
        dict,
        argparse.Namespace
    ]
) -> List[Batch]:
    """Worker function used to form batches from list of text

    Arguments:
        params {Tuple[List[str], dict, argparse.Namespace]} --
            Consists of the following
            - List of raw texts
            - Token dictionary
            - Parsed commandline options

    Returns:
        List[BATCH_TYPE] -- List of batches formed from raw texts.
    """
    chunk, dictionary, args = params
    batcher = load_batcher(dictionary, args)
    batches = list(batcher.make_batches(chunk))
    return batches


def append_results_to_file(
    embeddings: np.array,
    f: io.BufferedWriter,
    args: argparse.Namespace
):
    """Appends embeddings to output file

    Arguments:
        embeddings {np.array} -- Newly generated embeddings
        f {io.BufferedWriter} -- A bytes writier to store outputs
        args {argparse.Namespace} -- Parsed commandline options

    Raises:
        ValueError: output_format is invalid
    """
    output_dtype = np.float16 if args.output_fp16 else np.float32
    if embeddings.dtype != output_dtype:
        embeddings = embeddings.astype(output_dtype)

    if args.output_format == 'npy':
        embeddings.tofile(f)

    elif args.output_format == 'csv':
        import pandas as pd
        df = pd.DataFrame(embeddings)
        buffer = df.to_csv(header=False, index=False)
        assert buffer.endswith('\n')
        f.write(buffer.encode('utf-8'))

    else:
        msg = 'Unexpected output format "{}"'.format(args.output_format)
        raise ValueError(msg)


def main(args):
    if args.num_workers >= 1:
        import multiprocessing
        num_workers = args.num_workers
    else:
        from multiprocessing import dummy as multiprocessing
        num_workers = 1

    # Download pretrained model if needed
    download_pretrained_models(args)

    start_time = time()

    # Load encoder
    device = determine_device(args.gpu)
    encoder, dictionary = load_encoder_from_file(args.model)
    encoder = encoder.eval()
    encoder = encoder.to(device)
    if args.fp16:
        encoder = encoder.half()

    # Define create_loader fn
    semaphore = threading.Semaphore(num_workers)

    def create_loader():
        for chunk in gen_utils.chunk(
            create_input_stream(args),
            size=args.chunk_size,
        ):
            semaphore.acquire()
            yield chunk, dictionary, args

    # Define process_batches fn
    @torch.no_grad()
    def process_batches(batches: List[Batch]) -> Tuple[List[str], np.ndarray]:
        chunk_size = sum([len(batch[0]) for batch in batches])
        chunk_texts = [None] * chunk_size
        chunk_embeddings = [None] * chunk_size
        for b in batches:
            outputs = encoder.forward(
                b.tokens.to(device),
                b.lengths.to(device),
            ).sentemb
            outputs = outputs.detach().cpu().numpy()

            for i, text, embedding in zip(b.indices, b.texts, outputs):
                chunk_texts[i] = text
                chunk_embeddings[i] = embedding

        return chunk_texts, np.array(chunk_embeddings)

    # Remove existing output file if needed
    output_path = os.path.abspath(args.output)
    if os.path.isfile(output_path):
        os.remove(output_path)
    os.makedirs(os.path.dirname(output_path), exist_ok=True)

    # Loop through each chunk
    # 1. Create batches (on subprocesses)
    # 2. Inference with bilstm
    # 3. Save results to file
    with open(output_path, 'wb') as f:
        with tqdm(unit='sent') as pbar:
            with multiprocessing.Pool(num_workers) as pool:
                t0 = time()
                for batches in pool.imap(create_batches, create_loader()):
                    semaphore.release()

                    t1 = time()
                    _, embeddings = process_batches(batches)

                    t2 = time()
                    torch.cuda.empty_cache()
                    append_results_to_file(embeddings, f, args)

                    t3 = time()
                    pbar.set_postfix({
                        'data_time': t1 - t0,
                        'inference_time': t2 - t1,
                        'write_time': t3 - t2,
                    })
                    pbar.update(len(embeddings))
                    t0 = time()

    time_taken = time() - start_time
    print('Finished in {:.1f}s'.format(time_taken))
